"""

`oaaclient` utility functions.

Copyright 2022 Veza Technologies Inc.

Use of this source code is governed by the MIT
license that can be found in the LICENSE file or at
https://opensource.org/licenses/MIT.

helper functions commonly used by OAA integrations
"""
import base64
import json
import os
import logging

log = logging.getLogger(__name__)

def log_arg_error(log: object, arg: str = None, env: str = None) -> None:
    """Helper function for logging errors when loading parameters

    Helper function used to create consistent messages in connectors when required parameters can be set at command
    line or as environment variables.

    Message can include information on parameter and/or environment variable but must provide one.

    Args:
        log (object): logging facility object to log to
        arg (str, optional): Command line option for parameter such as `--veza-url`. Defaults to None.
        env (str, optional): OS Environment variable for parameter such as `VEZA_URL`. Defaults to None.

    Raises:
        Exception: if neither `arg` or `env` are supplied
    """

    if arg and env:
        log.error(f"Unable to load required parameter, must supply {arg} or set OS environment variable {env}")
    elif arg and not env:
        log.error(f"Unable to load required parameter, must supply {arg}")
    elif env:
        log.error(f"Unable to load required parameter, must set OS environment variable {env}")
    else:
        raise Exception("Must provide arg or env to include in error message")
    return


def load_json_from_file(json_path: str) -> dict:
    """Load JSON from file

    Args:
        json_path (str): path to JSON file on disk

    Raises:
        Exception: Unable to process JSON
        Exception: Error reading JSON file

    Returns:
        dict: JSON decoded to dictionary
    """
    try:
        with open(json_path) as f:
            data = json.load(f)
    except json.decoder.JSONDecodeError as e:
        raise Exception(f"Unable to process JSON from {json_path}: {e}")
    except OSError as e:
        raise Exception(f"Error reading file {json_path}: {e}")

    return data

def encode_icon_file(icon_path: str) -> str:
    """ read an icon file to a base64 encoded string

    Args:
        icon_path (str): Path to icon file on disk

    Returns:
        str: base64 encoding of file
    """

    with open(icon_path, "rb") as f:
        b64_icon = base64.b64encode(f.read())

    return b64_icon.decode()


def build_report(veza_con, report_definition: dict) -> dict:
    """Creates or updates a Veza report from a dictionary

    Creates a report and containing queries from a dictionary definition. Function will create any queries it does not
    find based on name. If a query with the same name already exists the existing query will be added to the report.

    If a report already exists with the same name any missing queries will be added to the report.

    `report_definition` must be a dictionary with `name` for the report and `queries` list of Veza query
    definitions:

        {"name": "My Report", "queries": [{..},{...}]}

    Args:
        veza_con (OAAClient): OAAClient connection to make Veza API calls
        report_definition (dict): Report definition

    Raises:
        ValueError: Missing name or queries key

    Returns:
        dict: API response from Report creation or update
    """

    report_name = report_definition.get("name")
    if not report_name:
        raise ValueError("Report source file must contain 'name'")

    if not report_definition.get("queries"):
        raise ValueError("Report source file must contain 'queries' list")

    # get all quires to know which queries need to be created and which don't
    all_queries = veza_con.get_queries()
    query_names = {}
    for q in all_queries:
        if q.get("query_type") == "SYSTEM_CREATED":
            continue
        query_names[q["name"]] = q["id"]

    # create queries that don't already exists
    query_ids = []
    for query in report_definition.get("queries", []):
        if query["name"] in query_names:
            log.debug(f"Found existing query with same name, using for report, {query['name']}")
            query_ids.append(query_names[query["name"]])
        else:
            response = veza_con.create_query(query=query)
            query_ids.append(response["id"])

    # get all reports to know if report already exists
    existing_reports = {}
    for e in veza_con.get_reports():
        id = e.get("id")
        name = e.get("name")
        existing_reports[name] = id

    response = {}
    if report_name not in existing_reports:
        # create a new report
        log.debug("Creating new report")
        report_definition = { "name": report_name, "description": report_name, "queries": []}
        for id in query_ids:
            report_definition["queries"].append({"query": id})
        response = veza_con.create_report(report=report_definition)
    else:
        # update existing report
        report_id = existing_reports[report_name]
        log.debug(f"Updating report {report_id}")
        for query_id in query_ids:
            response = veza_con.add_query_report(report_id=report_id, query_id=query_id)

    return response


import logging
import os
import sys


import pytest
from unittest.mock import MagicMock, patch

import oaaclient.client
import oaaclient.utils

# enable debug logging
logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger()
log.setLevel(logging.DEBUG)

@patch.object(oaaclient.client.OAAClient, "_test_connection", return_value=None)
@patch.object(os, "getenv", return_value="abc123")  # patch the getenv call that will be made for the Veza API key
@patch.object(sys, "argv", ["oaaclient-report-builder", "--host", "https://noreal.vezacloud.com", "tests/report_test.json"])
def test_report_builder_entrypoint(test_connection_mock, mock_getenv) -> None:

    with patch("oaaclient.utils.build_report") as mock_build_report:
        oaaclient.client.report_builder_entrypoint()

    assert mock_build_report.called
    mock_build_report.assert_called()



"""
Copyright 2022 Veza Technologies Inc.

Use of this source code is governed by the MIT
license that can be found in the LICENSE file or at
https://opensource.org/licenses/MIT.
"""

import logging
from logging import NullHandler

__version__ = "1.1.3"

# add null handler if calling class isn't logging
logging.getLogger(__name__).addHandler(NullHandler())



#!env python3
"""

Classes for calling Veza APIs and managing OAA providers and data sources.

Copyright 2022 Veza Technologies Inc.

Use of this source code is governed by the MIT
license that can be found in the LICENSE file or at
https://opensource.org/licenses/MIT.
"""

from __future__ import annotations

import argparse
import base64
import gzip
import json
import logging
import os
import platform
import re
import socket
import sys
import urllib.parse
from datetime import datetime
from enum import Enum

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import JSONDecodeError as RequestsJSONDecodeError
from urllib3.util.retry import Retry

from oaaclient import __version__ as OAACLIENT_VERSION
import oaaclient.utils as oaautils
from oaaclient.templates import CustomApplication, CustomIdPProvider

PROVIDER_ICON_MAX_SIZE = 64_000

log = logging.getLogger(__name__)

class OAAClientError(Exception):
    """Error raised by OAAClient.

    Raised for issues connecting to the OAA API and when the API returns an error.

    Args:
        error (str): short string for error message
        message (str): detailed error message
        status_code (int, optional): status code for HTTP related errors. Defaults to None.
        details (list, optional): list of additional details for error. Defaults to None.
    """

    def __init__(self, error: str, message:str , status_code: int = None, details: list = None) -> None:
        self.error = error
        self.message = message
        self.status_code = status_code
        if not details:
            self.details = []
        else:
            self.details = details

        super().__init__(self, f"{error}: {message}")

class OAAResponseError(OAAClientError):
    """ Error returned from API Call"""

    def __init__(self, *args, **kwargs):

        self.request_id = kwargs.pop("request_id", None)
        self.timestamp = kwargs.pop("timestamp", None)
        super().__init__(*args, **kwargs)
        return

class OAAConnectionError(OAAClientError):
    """ Error with API Connection"""

class OAAClient():
    """Class for OAA API Connection and Management.

    Utilities for OAA-related operations with Veza API calls. Manages custom providers and data sources, and can push OAA
    payloads from JSON or template objects.

    Connection url and API key can be automatically loaded from OS environment values if set. To utilize environment variables
    initialize OAAClient without providing a URL or API key value and set the `VEZA_URL` and `VEZA_API_KEY` OS environment variables.

    Args:
        url (str, optional): URL for Veza instance.
        api_key (str, optional): Veza API key.
        username (str, optional): Not used (legacy). Defaults to None.
        token (str, optional): Legacy parameter name for API key. Defaults to None.

    Attributes:
        url (str): URL of the Veza instance to connect to
        api_key (str): Veza API key
        enable_compression (bool): Enable or disable compression of the OAA payload during push, defaults to enabled (True)

    Raises:
        OAAClientError: For errors connecting to API and if API returns errors
    """
    # maximum number of times to retry a API calls, can be set with environment variable OAA_API_RETRIES
    DEFAULT_RETRY_COUNT = 10
    # Backoff multiplier factor for time between API retries in seconds
    # Back time exponential formula: {backoff factor} * (2 ^ ({retry number} - 1))
    DEFAULT_RETRY_BACKOFF_FACTOR = 0.6
    # maximum amount of time to backoff for between calls
    DEFAULT_RETRY_MAX_BACKOFF = 30

    # platform defined allowed characters for use unless otherwise allowed defined
    ALLOWED_CHARACTERS = r"^[ @#$%&*:()!,a-zA-Z0-9_'\"=.-]*$"

    # default number of entities to get when requesting a page size.
    # Must be passed with params for API call when calling paging APIs `params={"page_size": self.DEFAULT_PAGE_LENGTH}`
    DEFAULT_PAGE_SIZE = 250

    def __init__(self, url:str = None, api_key: str = None, username: str = None, token: str = None):

        if not url and "VEZA_URL" in os.environ:
            url = os.getenv("VEZA_URL", "")
        elif not url:
            raise ValueError("Must provide Veza URL")

        if not re.match(r"^https:\/\/.*", url):
            self.url = f"https://{url}"
        else:
            self.url = url
        self.url = self.url.rstrip("/")

        # for development purposes only sometimes system is run without signed certificates,
        # disable certificate verification only if VEZA_UNSAFE_HTTPS OS env variable is set to true
        self.verify_ssl = True
        unsafe_https = os.getenv("VEZA_UNSAFE_HTTPS", "")
        if unsafe_https.lower() == "true":
            self.verify_ssl = False

        self.username = username

        if api_key:
            self.api_key = api_key
        elif token:
            self.api_key = token
        elif "VEZA_API_KEY" in os.environ:
            self.api_key = os.getenv("VEZA_API_KEY", "")
        else:
            raise ValueError("Must provide Veza API key")

        if not self.url:
            raise OAAClientError("MISSING_URL", "URL cannot be None")
        if not self.api_key:
            raise OAAClientError("MISSING_AUTH", "API key cannot be None")

        # enable payload compression by default, connection object property can be set to False to disable
        self.enable_compression = True

        # setup retry strategy for API calls

        try:
            retry_count = int(os.getenv("OAA_API_RETRIES", self.DEFAULT_RETRY_COUNT))
        except ValueError:
            log.error(f"OAA_API_RETRIES variable must be integer, ignoring and setting to default {self.DEFAULT_RETRY_COUNT}")
            retry_count = self.DEFAULT_RETRY_COUNT

        retry_policy = OAARetry(backoff_max=self.DEFAULT_RETRY_MAX_BACKOFF, total=retry_count, backoff_factor=self.DEFAULT_RETRY_BACKOFF_FACTOR, status_forcelist=[429, 500, 502, 503, 504])
        adapter = HTTPAdapter(max_retries=retry_policy)
        self._http_adapter = requests.Session()
        self._http_adapter.mount("https://", adapter)
        self._http_adapter.mount("http://", adapter)

        self.update_user_agent()

        # test connection
        self._test_connection()

    def get_provider_list(self) -> list[dict]:
        """Return list of Providers.

        Returns:
            list[dict]: Returns a list of existing Providers as dictionaries
        """

        return self.api_get("/api/v1/providers/custom", params={"page_size": self.DEFAULT_PAGE_SIZE})

    def get_provider(self, name: str) -> dict:
        """Get Provider by name.

        Args:
            name (str): name of Provider

        Returns:
            dict: dictionary representing Provider or None
        """
        filter = {"filter": f"name eq \"{name}\""}
        response = self.api_get(f"/api/v1/providers/custom", params=filter)

        # expect a list of one or empty
        if not response:
            # no provider with that name found
            provider = None
        elif len(response) == 1:
            provider = response[0]
        else:
            # this shouldn't happen
            raise OAAClientError(error="Unexpected Results", message = "Unexpected results in response, returned more than one result")

        # return the provider detail
        return provider

    def get_provider_by_id(self, provider_id: str) -> dict:
        """Get Provider by UUID identifier.

        Args:
            provider_id (str): Unique global identifier for provider

        Returns:
            dict: dictionary representation of Provider or None
        """
        try:
            response = self.api_get(f"/api/v1/providers/custom/{provider_id}")
        except OAAResponseError as e:
            if e.status_code == 404:
                # Provider not found, return None
                return None
            else:
                raise e

        return response

    def create_provider(self, name: str, custom_template: str, base64_icon: str = None) -> dict:
        """Create a new Provider.

        Creates a new Provider with the given name. An error will be raised for naming conflicts.

        Args:
            name (str): new Provider name
            custom_template (str): the OAA template to use for the Provider ("application" or "identity_provider")
            base64_icon (str, optional): Base64 encoded string of icon to set for Provider. Defaults to None.

        Raises:
            ValueError: Provider name contains invalid characters

        Returns:
            dict: dictionary representing the created Provider
        """

        if not re.match(self.ALLOWED_CHARACTERS, name):
            raise ValueError(f"Provider name contains invalid characters, must match {self.ALLOWED_CHARACTERS}")

        provider = self.api_post("/api/v1/providers/custom", {"name": name, "custom_template": custom_template})
        if base64_icon:
            self.update_provider_icon(provider['id'], base64_icon)

        return provider

    def update_provider_icon(self, provider_id: str, base64_icon: str) -> None:
        """ Update an existing custom provider icon from base64 encoded string.

        To load an icon from file,  use `utils.encode_icon_file` to get the base64 encoding of the file first

        Args:
            provider_id (str): unique ID of existing provider
            base64_icon (str): base64 encoded string of new icon

        Raises:
            ValueError: If icon size exceeds maximum allowed size

        """
        if sys.getsizeof(base64_icon) > PROVIDER_ICON_MAX_SIZE:
            raise ValueError("Max icon size of 64KB exceeded")

        if isinstance(base64_icon, bytes):
            base64_icon = base64_icon.decode()

        icon_payload = {"icon_base64": base64_icon}
        self.api_post(f"/api/v1/providers/custom/{provider_id}:icon", data=icon_payload)

        return None

    def delete_provider(self, provider_id: str) -> dict:
        """Delete an existing provider by ID.

        Deleting a provider removes all datasources and historical data.
        Fully deleting the provider is a background operation that will complete after API response is returned.

        Args:
            provider_id (str): ID of provider to delete

        Returns:
            dict: API response
        """
        response = self.api_delete(f"/api/v1/providers/custom/{provider_id}")
        return response


    def get_data_sources(self, provider_id: str) -> list[dict]:
        """Get Data Sources for Provider by ID.

        Get the list of existing Data Sources, filtered by Provider UUID.

        Args:
            provider_id (str): ID of Provider

        Returns:
            list[dict]: List of Data Sources as dictionaries
        """

        return self.api_get(f"/api/v1/providers/custom/{provider_id}/datasources", params={"page_size": self.DEFAULT_PAGE_SIZE})


    def get_data_source(self, name:str, provider_id:str) -> dict:
        """Get Provider's Data Source by name.

        Find a Data Source from a specific provider based on the name of the Data Source

        Args:
            name (str): Data Source name
            provider_id (str): Provider unique ID

        Returns:
            dict: Data Source as dict or None
        """

        filter = {"filter": f"name eq \"{name}\""}
        response = self.api_get(f"/api/v1/providers/custom/{provider_id}/datasources", params=filter)

        # expect a list of one or empty
        if not response:
            # no data_source with that name found, return None
            data_source = None
        elif len(response) == 1:
            data_source = response[0]
        else:
            # this shouldn't happen
            raise OAAClientError(error="Unexpected Results", message = "Unexpected results in response, returned more than one result")

        return data_source

    def create_data_source(self, name: str, provider_id: str) -> dict:
        """Create a new Data Source for the given Provider ID.

        Args:
            name (str): Name for new Data Source
            provider_id (str): Unique identifier for the Provider

        Raises:
            ValueError: Data source name contains invalid characters

        Returns:
            dict: dictionary of new Data Source
        """

        if not re.match(self.ALLOWED_CHARACTERS, name):
            raise ValueError(f"Data source name contains invalid characters, must match {self.ALLOWED_CHARACTERS}")

        data_source = {"name": name, "id": provider_id}
        return self.api_post(f"/api/v1/providers/custom/{provider_id}/datasources", data=data_source)

    def create_datasource(self, name, provider_id):
        """
        **Deprecated** Legacy function for backward-compatibility.
        """
        return self.create_data_source(name, provider_id)

    def delete_data_source(self, data_source_id: str, provider_id: str) -> dict:
        """Delete an existing Data Source by ID.

        Removes a Data Source and all entity data.

        Args:
            data_source_id (str): ID of Data Source to delete
            provider_id (str): ID of Data Source Provider

        Returns:
            dict: API response
        """
        response = self.api_delete(f"/api/v1/providers/custom/{provider_id}/datasources/{data_source_id}")
        return response

    def push_metadata(self, provider_name: str, data_source_name: str, metadata: dict, save_json: bool = False) -> dict:
        """Push an OAA payload dictionary to Veza.

        Publishes the supplied `metadata` dictionary representing an OAA payload to the specified provider and
        data source. The function will create a new data source if it does not already exist, but requires the Provider be
        created ahead of time.

        Args:
            provider_name (str): Name of existing Provider
            data_source_name (str): Name for Data Source, will be created if doesn't exist.
            metadata (dict): Dictionary of OAA payload to push.
            save_json (bool, optional): Save the OAA JSON payload to a local file before push. Defaults to False.

        Raises:
            OAAClientError: If any API call returns an error including errors processing the OAA payload.

        Returns:
            dict: API response to the push request (including any warnings).
        """

        provider = self.get_provider(provider_name)
        if not provider:
            raise OAAClientError("NO_PROVIDER", f"Unable to locate provider {provider_name}, cannot push without existing provider")
        data_source = self.get_data_source(data_source_name, provider["id"])
        if not data_source:
            self.create_datasource(data_source_name, provider["id"])
            data_source = self.get_data_source(data_source_name, provider["id"])

        if save_json:
            ts = datetime.now().strftime("%Y%m%d-%H%M%S")
            out_name = f"{data_source_name}-{ts}.json"
            with open(out_name, "w") as f:
                f.write(json.dumps(metadata, indent=2))

        if self.enable_compression:
            log.debug("Compressing payload")
            metadata_bytes = json.dumps(metadata).encode()
            metadata_size = sys.getsizeof(metadata_bytes)
            compressed_bytes = gzip.compress(metadata_bytes)
            del metadata_bytes

            encoded = base64.b64encode(compressed_bytes).decode()
            encoded_size = sys.getsizeof(encoded)
            del compressed_bytes
            log.debug(f"Compression complete, payload size in bytes: {metadata_size:,}, encoded compressed: {encoded_size:,}")
            payload = {"id": provider["id"], "data_source_id": data_source["id"], "json_data": encoded, "compression_type": "GZIP"}
        else:
            payload = {"id": provider["id"], "data_source_id": data_source["id"], "json_data": json.dumps(metadata)}

        payload_size = sys.getsizeof(payload["json_data"])
        if payload_size > 100_000_000:
            raise OAAClientError("OVERSIZE", message=f"Payload size exceeds maximum size of 100MB: {payload_size:,} bytes, compression enabled: {self.enable_compression}")

        log.debug(f"Final payload size: {payload_size:,} bytes")
        result = self.api_post(f"/api/v1/providers/custom/{provider['id']}/datasources/{data_source['id']}:push", payload)

        return result

    def push_application(self, provider_name: str, data_source_name: str, application_object: CustomApplication|CustomIdPProvider, save_json=False) -> dict:
        """Push an OAA Application Object (such as CustomApplication).

        Extracts the OAA JSON payload from the supplied OAA class (e.g. CustomApplication, CustomIdPProvider) and push to
        the supplied Data Source.

        The Provider must be a valid Provider (created ahead of time). A new data source will be created
        if it does not already exist.

        For logging, and debugging, the optional `save_json` flag writes the payload to a local file (before push). Output file name
        is formatted with a timestamp: `{data source name}-{%Y%m%d-%H%M%S}.json`

        Args:
            provider_name (str): Name of an existing Provider.
            data_source_name (str): Name for Data Source (will be created if it doesn't exist).
            application_object (Class): OAA object to extract the payload from
            save_json (bool, optional): Save the JSON payload to a local file before push. Defaults to False.

        Raises:
            OAAClientError: If any API call returns an error (including errors processing the OAA payload).

        Returns:
            dict: API response from push, including any warnings that are returned.
        """
        metadata = application_object.get_payload()
        return self.push_metadata(provider_name, data_source_name, metadata, save_json=save_json)

    def get_queries(self, include_inactive_queries:bool = True) -> list[dict]:
        """Get all saved Assessment Queries

        Veza can filter out queries that include inactive entity types (e.g. Okta Users on a system without Okta configured).
        To only retrieve queries that include active entity types set `include_inactive_queries` to False.

        Args:
            include_inactive_queries (bool): Set False to exclude inactive queries from result. Defaults to True.

        Returns:
            list: List of assessment Queries as dictionaries
        """

        params = {}
        params["include_inactive_queries"] = include_inactive_queries
        params["page_size"] = self.DEFAULT_PAGE_SIZE

        return self.api_get("/api/v1/assessments/queries", params=params)

    def get_query_by_id(self, id: str) -> dict:
        """Get Assessment Query by ID

        Args:
            id (str): UUID identifier for Query

        Returns:
            dict: Query definition
        """

        return self.api_get(f"/api/v1/assessments/queries/{id}")

    def create_query(self, query: dict) -> dict:
        """Create a new Assessment Query

        For details on how to define an Assessment Query see the Veza docs.

        Args:
            query (dict): Query definition

        Returns:
            dict: API response including ID of created Query
        """

        return self.api_post("/api/v1/assessments/queries", data=query)

    def delete_query(self, id: str, force: bool = False) -> dict:
        """Delete an Assessment Query by ID

        Args:
            id (str): UUID of Query to delete
            force (bool): Force deletion of query that may be part of a report. Defaults to False

        Returns:
            dict: API response from delete
        """

        params = {"force": force}
        return self.api_delete(f"/api/v1/assessments/queries/{id}", params)

    def get_reports(self, include_inactive_reports: bool = True, include_inactive_queries: bool = True) -> list[dict]:
        """Get all Reports

        Gets Reports created on the system. To get all reports `include_inactive_reports` and `include_inactive_queries` must be set
        to True.

        Args:
            include_inactive_reports (bool, Optional): Set to True to include reports that contain no active providers, defaults to True.
            include_inactive_queries (bool, Optional): Set to True to include reports that contain only inactive queries, defaults to True.

        Returns:
            list[dict]: List of Reports as dictionary objects
        """

        params = {}
        params["include_inactive_reports"] = include_inactive_reports
        params["include_inactive_queries"] = include_inactive_queries
        params["page_size"] = self.DEFAULT_PAGE_SIZE

        return self.api_get("/api/preview/assessments/reports", params=params)

    def get_report_by_id(self, id: str, include_inactive_queries: bool = True) -> dict:
        """Get Report by ID

        Veza can filter out queries from reports that only contain entity types that are not configured (e.g. Okta
        Users on a system without Okta configured). To only return queries configured on the report that match entity
        types configured on the system set `include_inactive_queries` to `False`

        Args:
            id (str): UUID of Report to get
            include_inactive_queries (bool): Set True to include inactive queries. Default True.
        Returns:
            dict: Report definition
        """

        params = {"include_inactive_queries": include_inactive_queries}
        return self.api_get(f"/api/preview/assessments/reports/{id}", params=params)

    def create_report(self, report: dict) -> dict:
        """Create a new Report

        For details on how to define a new Report see the Veza docs.

        Args:
            report (dict): Report definition

        Returns:
            dict: API response including ID of created Report
        """

        return self.api_post("/api/preview/assessments/reports", data=report)

    def update_report(self, report_id: str, report: dict) -> dict:
        """Update an existing report

        Args:
            report_id (str): UUID of Report to updated
            report (dict): Updated Report definition

        Returns:
            dict: API response
        """

        return self.api_put(f"/api/preview/assessments/reports/{report_id}", data=report)

    def add_query_report(self, report_id: str, query_id: str) -> dict:
        """Add a Query to a Report

        Adds a Query to an existing Report by ID

        Args:
            report_id (str): Report UUID to add Query to
            query_id (str): Query UUID to add

        Returns:
            dict: API Response
        """

        return self.api_put(f"/api/preview/assessments/reports/{report_id}/queries/{query_id}", data=None)

    def delete_report(self, id: str) -> dict:
        """Delete Report by ID

        Args:
            id (str): UUID of Report to delete

        Returns:
            dict: API response
        """

        return self.api_delete(f"/api/preview/assessments/reports/{id}")

    def api_get(self, api_path: str, params: dict = None) -> list|dict:
        """Perform a Veza API GET operation.

        Makes the GET API call to the given path and processes the API response. Returns the `value` or `values` result
        returned from the API.

        - For API endpoints that return a list like `/api/v1/providers/custom` function will return a list of entities or an
        empty list if the API returns no results.
        - For API endpoints that are a specific ID such as `/api/v1/providers/custom/<uuid>` function will return the
        dictionary result of the JSON returned by the API.

        Args:
            api_path (str): API path relative to Veza URL (example `/api/v1/providers`).

        Raises:
            OAAResponseError: API returned an error
            OAAConnectionError: Connection error during HTTP operation

        Returns:
            list|dict: Returns list or dict based on API destination
        """

        result = []

        if not params:
            params = {}

        while True:
            response = self._perform_request(method="GET", api_path=api_path, params=params)


            if "values" in response:
                # paginated response,
                result.extend(response.get("values", []))
                if response.get("has_more"):
                    params["page_token"] = response.get("next_page_token")
                else:
                    # no more pages
                    break
            elif "value" in response:
                # singular API response, set return value to response value
                result = response["value"]
                break
            else:
                # unexpected API response, just return the result
                return response

        return result

    def api_post(self, api_path: str, data: dict, params: dict = None) -> list|dict:
        """Perform a Veza API POST operation.

        Call POST on the supplied Veza instance API path, submitting a data payload.

        Returns `value` or `values` response from API result. Paginated responses are automatically processed to
        collect all responses a single list.

        Args:
            api_path (str): API path relative to Veza URL example `/api/v1/providers`
            data (dict): dictionary object included as JSON in body of POST operation
            params (dict, optional): Optional HTTP query parameters. Defaults to empty dictionary.

        Raises:
            OAAResponseError: API returned an error
            OAAConnectionError: Connection error during HTTP operation

        Returns:
            dict: API response as dictionary
        """

        if not params:
            params = {}

        result = []
        while True:
            response = self._perform_request(method="POST", api_path=api_path, data=data, params=params)


            if "values" in response:
                # paginated response,
                result.extend(response.get("values", []))
                if response.get("has_more"):
                    params["page_token"] = response.get("next_page_token")
                else:
                    # no more pages
                    break
            elif "value" in response:
                # singular API response, set return value to response value
                result = response["value"]
                break
            else:
                # unexpected API response, just return the result
                return response

        return result

    def api_put(self, api_path: str, data: dict, params: dict = None) -> list|dict:
        """Perform Veza API PUT operation.

        Call PUT on the supplied Veza instance API path, including the data payload.

        Returns `value` or `values` response from API result. Paginated responses are automatically processed to
        collect all responses a single list.

        Args:
            api_path (str): API path relative to Veza URL example `/api/v1/providers`
            data (dict): dictionary object included as JSON in body of PUT operation
            params (dict, optional): Optional HTTP query parameters. Defaults to empty dictionary.

        Raises:
            OAAResponseError: API returned an error
            OAAConnectionError: Connection error during HTTP operation

        Returns:
            dict: API response as dictionary
        """

        if not params:
            params = {}

        result = []
        while True:
            response = self._perform_request(method="PUT", api_path=api_path, data=data, params=params)

            if "values" in response:
                # paginated response,
                result.extend(response.get("values", []))
                if response.get("has_more"):
                    params["page_token"] = response.get("next_page_token")
                else:
                    # no more pages
                    break
            elif "value" in response:
                # singular API response, set return value to response value
                result = response["value"]
                break
            else:
                # unexpected API response, just return the result
                return response

        return result

    def api_delete(self, api_path:str, params: dict = None) -> dict:
        """Perform REST API DELETE operation.

        Args:
            api_path (str): API Path API path relative to Veza URL

        Raises:
            OAAResponseError: API returned an error
            OAAConnectionError: Connection error during HTTP operation

        Returns:
            dict: API response from call
        """
        return self._perform_request(method="DELETE", api_path=api_path, params=params)

    def _perform_request(self, method: str, api_path: str, data: dict = None, params: dict = None) -> dict:
        """Perform HTTP request

        Performs an HTTP request of the specified method to the Veza tenant

        Args:
            method (str): HTTP method, GET, POST, DELETE
            api_path (str): API path relative to Veza host URL, e.g. `/api/v1/providers`
            data (dict, optional): For POST operation data to send. Defaults to None.

        Raises:
            OAAClientError: For errors connecting to or returned by the Veza tenant

        Returns:
            dict: Veza API JSON response as dictionary
        """

        response = None

        headers = {}
        headers["authorization"] = f"Bearer {self.api_key}"
        headers["user-agent"] = self._user_agent
        api_timeout = 300
        api_path = api_path.lstrip("/")

        if params:
            params_str = urllib.parse.urlencode(params)
        else:
            params_str = None

        try:
            response = self._http_adapter.request(method, f"{self.url}/{api_path}", headers=headers, timeout=api_timeout, params=params_str, json=data, verify=self.verify_ssl)
            response.raise_for_status()
            # load API response
            result = response.json()
        except requests.exceptions.HTTPError as e:
            # HTTP request completed but returned an error
            # decode expected error message parts
            details = []
            timestamp = None
            request_id = None
            try:
                result = response.json()
                message = result.get("message", f"Unknown error during {method.upper()}")
                code = result.get("code", "UNKNOWN")
                timestamp = result.get("timestamp", None)
                request_id = result.get("request_id", None)
                details = result.get("details", [])
            except requests.exceptions.JSONDecodeError:
                # response is not a valid JSON, unexpected
                result = {}
                code = "ERROR"
                if response.reason:
                    message = f"Error reason: {response.reason}"
                else:
                    message = "Unknown error, response is not JSON"

            log.debug(f"Error returned by Veza API: {e.response.status_code} {message} {e.response.url} request_id: {request_id} timestamp {timestamp}")
            for d in details:
                log.debug(d)

            raise OAAResponseError(code, message, status_code=e.response.status_code, details=details, timestamp=timestamp, request_id=request_id)
        except requests.exceptions.JSONDecodeError as e:
            # HTTP response reports success but response does not decode to JSON
            if response:
                status_code = response.status_code
            else:
                status_code = None
            raise OAAClientError("ERROR", "Response not JSON", status_code=status_code)
        except requests.exceptions.RequestException as e:
            if not e.response:
                raise OAAConnectionError("ERROR", message=str(e))
            else:
                raise OAAConnectionError("ERROR", message=str(e), status_code=e.response.status_code)

        return result

    def _test_connection(self) -> None:
        """Test the OAAClient connection

        Validates the Veza URL and API key with a test call. Raises an exception
        when unable to connect to the Veza host.

        Raises:
            OAAResponseError: For errors returned by Veza including 401 for bad credentials
            OAAConnectionError: When unable to connect to Veza API.
        """


        # do separate DNS lookup to quickly fail if the hostname isn't resolving
        veza_url = urllib.parse.urlsplit(self.url)
        try:
            x = socket.getaddrinfo(veza_url.hostname, 0)
        except socket.gaierror as e:
            log.debug("DNS lookup failure during connection test")
            raise OAAConnectionError("Unknown host", message=f"Unable to lookup DNS for {veza_url.hostname}")

        # test connection to validate host and credentials
        try:
            self.api_get("/api/v1/providers/custom/templates")
        except OAAResponseError as e:
            log.debug("API returned error during API connection test")
            raise e
        except OAAConnectionError as e:
            log.debug("Unable to connect during API connection test")
            raise e

    def update_user_agent(self, extra: str = "") -> None:
        """Updates the User-Agent string passed with all API calls

        Generates a User-Agent with the oaaclient version, Python version and platform information.

        The optional `extra` string will be appended if provided.

        Args:
            extra (str, optional): Additional information to append to User-Agent string. Defaults to "".
        """

        python_version = platform.python_version()
        os_name = platform.system()
        os_version = platform.release()

        self._user_agent = f"oaaclient/{OAACLIENT_VERSION} python/{python_version} {os_name}/{os_version};"
        if extra:
            self._user_agent += f" {extra}"

        log.debug(f"User-Agent {self._user_agent}")

        return

class OAARetry(Retry):
    """Super class for urllib3.util.retry

    Super class to allow modifying the default max backoff time from 120 seconds
    to our own value

    Args:
        Retry (_type_): _description_
    """
    def __init__(self, backoff_max=30, **kwargs) -> None:
        super(OAARetry, self).__init__(**kwargs)
        #urllib3 1.x and 2.x have slightly different behavior for BACKOFF_MAX, setting both values covers both
        self.DEFAULT_BACKOFF_MAX = backoff_max
        self.backoff_max = backoff_max


def report_builder_entrypoint() -> None:
    """Entry point for `oaaclient-report-builder` command

    Reads a JSON file and passes it to the `oaaclient.utils.build_report` method
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", default=os.getenv("VEZA_URL"), help="URL endpoint for Veza Deployment")
    parser.add_argument("report_file", help="Path to source report file")
    args = parser.parse_args()

    # setup a log handler
    log = logging.getLogger()
    logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.INFO)

    report_file = args.report_file
    if not os.path.isfile(report_file):
        log.error(f"Unable to locate source file at path {report_file}")
        sys.exit(1)
    try:
        report_definition = oaautils.load_json_from_file(report_file)
    except Exception as e:
        log.error(e)
        sys.exit(1)

    veza_url = args.host
    veza_api_key = os.getenv("VEZA_API_KEY", "")
    if not veza_url:
        oaautils.log_arg_error(log, "--host", "VEZA_URL")
    if not veza_api_key:
        oaautils.log_arg_error(log, None, "VEZA_API_KEY")
    if not veza_url and veza_api_key:
        sys.exit(1)

    vezacon = OAAClient(veza_url, api_key=veza_api_key)

    log.info(f"Loading report from {report_file}")
    try:
        oaautils.build_report(vezacon, report_definition=report_definition)
    except OAAResponseError as e:
        log.error("Veza API error encounter building report")
        log.error(f"OAAResponseError {e.message} {e.status_code}")
        log.error(e.details)
        sys.exit(1)
    except OAAConnectionError as e:
        log.error("Error connecting to Veza")
        log.error(f"OAAConnectionError {e.message} {e.status_code}")
        sys.exit(1)
    except Exception as e:
        log.error("Error building report")
        log.error(e)
        sys.exit(1)

    log.info("Finished")
    return


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--auth-file", help="Config file with authentication information")
    parser.add_argument("--host", help="URL endpoint for Veza Deployment")
    parser.add_argument("--user", help="Username to connect as")
    parser.add_argument("--provider", required=True, help="Provider definition json file, will create if doesn't exist")
    parser.add_argument("metadata", help="Metadata json file to push, uses name of file as datasource name")
    args = parser.parse_args()

    host = None
    user = None
    password = None
    token = None

    if args.auth_file:
        if not os.path.isfile(args.auth_file):
            print(f"Error: Unable to locate auth file {args.auth_file}", file=sys.stderr)
            sys.exit(1)
        auth_config = oaautils.load_json_from_file(args.auth_file)

        host = auth_config['host']
        user = auth_config['user']
        if "token" in auth_config:
            token = auth_config['token']
        elif "password" in auth_config:
            password = auth_config['password']
    else:
        host = args.host
        user = args.user

    provider_metadata = oaautils.load_json_from_file(args.provider)

    try:
        provider_name = provider_metadata["name"]
        custom_template = provider_metadata['custom_template']
    except KeyError as e:
        raise Exception(f"Missing value in app template: {e}")

    try:
        con = OAAClient(host, api_key=token)
        provider = con.get_provider(provider_name)
        if provider:
            print("-- Found existing provider")
        else:
            print(f"++ Creating Provider {provider_name} of type {custom_template}")
            provider = con.create_provider(provider_name, custom_template)
        print(f"-- Provider: {provider['name']} ({provider['id']})")
        # utilize the file name as the datasource name
        data_source_name = os.path.splitext(os.path.basename(args.metadata))[0]

        print("-- Pushing metadata")
        metadata = oaautils.load_json_from_file(args.metadata)
        response = con.push_metadata(provider_name, data_source_name, metadata)
        if response.get("warnings", None):
            print("-- Push succeeded with warnings:")
            for e in response["warnings"]:
                print(f"  - {e}")
    except OAAClientError as e:
        print(f"-- Error: {e.error}: {e.message} ({e.status_code})", file=sys.stderr)
        if hasattr(e, "details"):
            for d in e.details:
                print(f"  -- {json.dumps(d, indent=2)}")
    return


if __name__ == "__main__":
    main()



"""
Copyright 2022 Veza Technologies Inc.

Use of this source code is governed by the MIT
license that can be found in the LICENSE file or at
https://opensource.org/licenses/MIT.
"""

import pytest

from generate_app import generate_app
from generate_idp import generate_idp
from oaaclient.templates import Tag

def test_custom_app_reprs() -> None:
    # quick test to ensure that all the str and repr functions are free of bugs

    app = generate_app()

    # app
    assert app.__str__() == "Custom Application pytest generated app - pytest"
    assert app.__repr__() == "CustomApplication(name='pytest generated app', application_type='pytest', description='This is a test')"

    # test local user
    bob = app.local_users["bob"]
    assert bob.__str__() == "Local User - bob (None)"
    assert bob.__repr__() == "LocalUser(name='bob', unique_id=None, identities=['bob@example.com'])"

    group1 = app.local_groups["group1"]
    assert group1.__str__() == "Local Group - group1 (None)"
    assert group1.__repr__() == "LocalGroup(name='group1', unique_id=None, identities=None)"

    role1 = app.local_roles["role1"]
    assert role1.__str__() == "Local Role - role1 (None)"
    assert role1.__repr__() == "LocalRole(name='role1', permissions=['all', 'Admin', 'Manage_Thing'], unique_id=None)"

    all_permission = app.custom_permissions["all"]
    assert all_permission.__str__() == "Custom Permission all - [OAAPermission.DataRead, OAAPermission.DataWrite, OAAPermission.DataCreate, OAAPermission.DataDelete, OAAPermission.MetadataRead, OAAPermission.MetadataWrite, OAAPermission.MetadataCreate, OAAPermission.MetadataDelete, OAAPermission.NonData]"
    assert all_permission.__repr__() == "CustomPermissions(name='all', permissions=[OAAPermission.DataRead, OAAPermission.DataWrite, OAAPermission.DataCreate, OAAPermission.DataDelete, OAAPermission.MetadataRead, OAAPermission.MetadataWrite, OAAPermission.MetadataCreate, OAAPermission.MetadataDelete, OAAPermission.NonData], apply_to_sub_resources=False)"

    thing1 = app.resources["thing1"]
    assert thing1.__str__() == "Resource: thing1 (None) - thing"
    assert thing1.__repr__() == "CustomResource(name='thing1', resource_type='thing', unique_id=None, application_name='pytest generated app', resource_key='thing1')"

    idp_user = app.idp_identities["user01@example.com"]
    assert idp_user.__str__() == "IdP Identity user01@example.com"
    assert idp_user.__repr__() == "IdPIdentity(name='user01@example.com')"

    assert app.property_definitions.__str__() == "ApplicationPropertyDefinitions for pytest"
    assert app.property_definitions.__repr__() == "ApplicationPropertyDefinitions(application_type='pytest')"

def test_custom_idp_reprs() -> None:

    idp = generate_idp()

    assert idp.__str__() == "Custom IdP Provider Pytest IdP - pytest"
    assert idp.__repr__() == "CustomIdPProvider(name='Pytest IdP', idp_type='pytest', domain=CustomIdPDomain(name='example.com'), description='Pytest Test IdP')"

    assert idp.domain.__str__() == "Custom IdP Domain example.com"
    assert idp.domain.__repr__() == "CustomIdPDomain(name='example.com')"

    user1 = idp.users["0001"]
    assert user1.__str__() == "IdP User - user0001 (0001)"
    assert user1.__repr__() == "CustomIdPUser(name='user0001', email='user001@example.com', full_name='User 0001', identity='0001')"

    group1 = idp.groups["g001"]
    assert group1.__str__() == "IdP Group group001 (g001)"
    assert group1.__repr__() == "CustomIdPGroup(name='group001', full_name='Group 001', identity='g001')"

    assert idp.property_definitions.__str__() == "IdP Property Definitions"
    assert idp.property_definitions.__repr__() == "IdPPropertyDefinitions()"

def test_tag_repr() -> None:

    test_tag = Tag("test", "value")
    assert test_tag.__str__() == "Tag test:value"
    assert test_tag.__repr__() == "Tag(key='test', value='value')"

    test_tag = Tag("Test")
    assert test_tag.__str__() == "Tag Test"
    assert test_tag.__repr__() == "Tag(key='Test', value='')"


"""
Copyright 2023 Veza Technologies Inc.

Use of this source code is governed by the MIT
license that can be found in the LICENSE file or at
https://opensource.org/licenses/MIT.
"""

from collections.abc import Mapping, MutableMapping

class CaseInsensitiveDict(MutableMapping):
    """Case Insensitive Key Dictionary

    Dictionary like object with case insensitive keys for types that support `.lower()` such as strings.

    Keys do not have to be strings, in the case where the key type does not support `.lower()` such as integers the
    value is used as is.

    Example:
        >>> from oaaclient.structures import CaseInsensitiveDict
        >>> x = CaseInsensitiveDict()
        >>> x["User"] = "value"
        >>> x.get("user")
        'value'
        >>> "USER" in x
        True
        >>> print(x)
        {'user': 'value'}
        >>> x
        CaseInsensitiveDict({'user': 'value'})

    """

    def __init__(self, data=None, **kwargs) -> None:
        self._entities = dict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __getitem__(self, key):
        try:
            key = key.lower()
        except AttributeError:
            pass

        return self._entities[key][1]

    def __setitem__(self, key, value) -> None:
        try:
            key = key.lower()
        except AttributeError:
            pass

        self._entities[key] = (key, value)

    def __delitem__(self, key) -> None:
        try:
            key = key.lower()
        except AttributeError:
            pass

        del self._entities[key]

    def __iter__(self):
        return (key for key, value in self._entities.values())

    def __len__(self) -> int:
        return len(self._entities)

    def __eq__(self, other):
        return NotImplemented

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({dict(self.items())})"

    def __str__(self) -> str:
        return f"{dict(self.items())}"


"""
Copyright 2022 Veza Technologies Inc.

Use of this source code is governed by the MIT
license that can be found in the LICENSE file or at
https://opensource.org/licenses/MIT.
"""

import os
import pytest
import re
import time

from oaaclient.client import OAAClient, OAAClientError
import oaaclient.utils as utils

from generate_app import generate_app
from generate_idp import generate_idp
from generate_app_id_mapping import generate_app_id_mapping

# set the timeout for the push tests, if the the datasource does not parse
TEST_TIMEOUT = os.getenv("OAA_PUSH_TIMEOUT", 300)

# REGEX for matching failure data source parsing messages for fast failing tests
FAILURE_REGEX = re.compile(r"fail|error", re.IGNORECASE)

@pytest.mark.skipif(not os.getenv("PYTEST_VEZA_HOST"), reason="Test host is not configured")
@pytest.mark.timeout(TEST_TIMEOUT)
def test_payload_push(veza_con, app_provider):
    # make sure compression is disabled
    veza_con.enable_compression = False
    app = generate_app()

    data_source_name = os.environ.get('PYTEST_CURRENT_TEST').replace("/", "-")

    b64_icon = utils.encode_icon_file("tests/oaa_icon.png")
    veza_con.update_provider_icon(provider_id=app_provider['id'], base64_icon=b64_icon)

    response = veza_con.push_application(app_provider['name'],
                                           data_source_name=data_source_name,
                                           application_object=app
                                           )
    if not response:
        assert False

    # Veza API always returns the warnings key, the list may be empty, in this case we expect it not to be
    assert "warnings" in response
    # since our payload includes fake identities expect warnings about not matching identities
    assert response["warnings"] is not None
    assert len(response["warnings"]) == 6
    for warning in response["warnings"]:
        if "Role is missing permission" in warning.get("message", ""):
            pass
        elif "Cannot find identity by names" in warning.get("message", ""):
            pass
        else:
            assert False, "Got unexpected warning from response"

    data_source = veza_con.get_data_source(data_source_name, provider_id=app_provider["id"])
    while True:
        data_source = veza_con.get_data_source(data_source_name, provider_id=app_provider["id"])
        if data_source["status"] == "SUCCESS":
            break
        elif FAILURE_REGEX.match(data_source["status"]):
            print(data_source)
            assert False, "Datasource parsing failure"
        else:
            time.sleep(4)


@pytest.mark.skipif(not os.getenv("PYTEST_VEZA_HOST"), reason="Test host is not configured")
@pytest.mark.timeout(TEST_TIMEOUT)
def test_payload_push_compressed(veza_con, app_provider):
    # enable compression
    veza_con.enable_compression = True
    app = generate_app()
    data_source_name = os.environ.get('PYTEST_CURRENT_TEST').replace("/", "-")

    b64_icon = utils.encode_icon_file("tests/oaa_icon.png")
    veza_con.update_provider_icon(provider_id=app_provider['id'], base64_icon=b64_icon)

    response = veza_con.push_application(app_provider['name'],
                                           data_source_name=data_source_name,
                                           application_object=app
                                           )
    if not response:
        assert False

    # Veza API always returns the warnings key, the list may be empty, in this case we expect it not to be
    assert "warnings" in response
    # since our payload includes fake identities expect warnings about not matching identities
    assert response["warnings"] is not None
    assert len(response["warnings"]) == 6
    for warning in response["warnings"]:
        if "Role is missing permission" in warning.get("message", ""):
            pass
        elif "Cannot find identity by names" in warning.get("message", ""):
            pass
        else:
            assert False, "Got unexpected warning from response"

    data_source = veza_con.get_data_source(data_source_name, provider_id=app_provider["id"])
    while True:
        data_source = veza_con.get_data_source(data_source_name, provider_id=app_provider["id"])
        if data_source["status"] == "SUCCESS":
            break
        elif FAILURE_REGEX.match(data_source["status"]):
            print(data_source)
            assert False, "Datasource parsing failure"
        else:
            time.sleep(4)



@pytest.mark.skipif(not os.getenv("PYTEST_VEZA_HOST"), reason="Test host is not configured")
@pytest.mark.timeout(TEST_TIMEOUT)
def test_payload_push_id_mapping(veza_con, app_provider):
    """ test for app payload where identities are mapped by id instead of name """

    app = generate_app_id_mapping()
    data_source_name = os.environ.get('PYTEST_CURRENT_TEST').replace("/", "-")

    b64_icon = utils.encode_icon_file("tests/oaa_icon.png")
    veza_con.update_provider_icon(provider_id=app_provider['id'], base64_icon=b64_icon)

    response = None
    try:
        response = veza_con.push_application(app_provider['name'],
                                            data_source_name=data_source_name,
                                            application_object=app
                                            )
    except OAAClientError as e:
        print(e)
        print(e.details)
        assert False

    if not response:
        assert False

    # Veza API always returns the warnings key, the list may be empty, in this case we expect it not to be
    assert "warnings" in response
    # since our payload includes fake identities expect warnings about not matching identities
    assert response["warnings"] is not None
    for warning in response["warnings"]:
        assert warning['message'].startswith("Cannot find identity by names")

    data_source = veza_con.get_data_source(data_source_name, provider_id=app_provider["id"])
    while True:
        data_source = veza_con.get_data_source(data_source_name, provider_id=app_provider["id"])
        if data_source["status"] == "SUCCESS":
            break
        elif FAILURE_REGEX.match(data_source["status"]):
            print(data_source)
            assert False, "Datasource parsing failure"
        else:
            time.sleep(4)


@pytest.mark.skipif(not os.getenv("PYTEST_VEZA_HOST"), reason="Test host is not configured")
def test_bad_payload(veza_con, app_provider):

    app = generate_app()
    payload = app.get_payload()
    # break the payload so it will throw an error
    payload['applications'][0]["bad_property"] = "This will break things"

    data_source_name = os.environ.get('PYTEST_CURRENT_TEST').replace("/", "-")

    with pytest.raises(OAAClientError) as e:
        response = veza_con.push_metadata(provider_name=app_provider['name'], data_source_name=data_source_name, metadata=payload)

    assert e.value.message is not None
    assert e.value.details is not None
    assert e.value.status_code == 400

    return


@pytest.mark.skipif(not os.getenv("PYTEST_VEZA_HOST"), reason="Test host is not configured")
@pytest.mark.timeout(TEST_TIMEOUT)
def test_idp_payload_push(veza_con, idp_provider):

    idp = generate_idp()

    data_source_name = os.environ.get('PYTEST_CURRENT_TEST').replace("/", "-")
    response = veza_con.push_application(idp_provider['name'],
                                           data_source_name=data_source_name,
                                           application_object=idp
                                           )
    if not response:
        assert False

    # print out any warnigns from the push for debugging puproses, warnings are not a failure, warnings will include
    # being unable to find fake identities
    if response.get("warnings", None):
        print("Push warnings:")
        for e in response["warnings"]:
            print(f"  - {e}")

    while True:
        data_source = veza_con.get_data_source(data_source_name, provider_id=idp_provider["id"])
        if data_source["status"] == "SUCCESS":
            break
        elif FAILURE_REGEX.match(data_source["status"]):
            print(data_source)
            assert False, "Datasource parsing failure"
        else:
            time.sleep(4)

    return



"""
Copyright 2022 Veza Technologies Inc.

Use of this source code is governed by the MIT
license that can be found in the LICENSE file or at
https://opensource.org/licenses/MIT.
"""

import pytest

from oaaclient.structures import CaseInsensitiveDict

def test_caseinsensitivedict():

    x = CaseInsensitiveDict()

    x["User"] = "value"
    assert x["User"] == "value"
    assert x["user"] == "value"
    assert x.get("USER") == "value"

    assert "user" in x
    assert "user" in x.keys()

    x = CaseInsensitiveDict()
    x["User1"] = "value1"
    x["User2"] = "value2"
    x["user3"] = "value3"

    assert len(x) == 3
    x["user1"] = "value1_updated"
    assert len(x) == 3
    assert x["User1"] == "value1_updated"
    assert x["user1"] == "value1_updated"

    assert x.__str__() == "{'user1': 'value1_updated', 'user2': 'value2', 'user3': 'value3'}"
    assert x.__repr__() == "CaseInsensitiveDict({'user1': 'value1_updated', 'user2': 'value2', 'user3': 'value3'})"
    assert f"{x}" == "{'user1': 'value1_updated', 'user2': 'value2', 'user3': 'value3'}"

    # test iterators
    expected_keys = ["user1", "user2", "user3"]
    expected_values = ["value1_updated", "value2", "value3"]

    # test basic iter
    received_keys = []
    for k in x:
        received_keys.append(k)

    # keys will all come back lowercased
    assert received_keys == expected_keys

    # test .keys() iter
    received_keys = []
    for k in x.keys():
        received_keys.append(k)

    # keys will all come back lowercased
    assert received_keys == expected_keys

    # test .values iter
    received_values = []
    for v in x.values():
        received_values.append(v)

    assert received_values == expected_values

    # test .items iter
    received_keys = []
    received_values = []
    for k, v in x.items():
        received_keys.append(k)
        received_values.append(v)

    assert received_keys == expected_keys
    assert received_values == expected_values

    # test for mixed types
    mixed = CaseInsensitiveDict()

    mixed["string"] = "I am a string"
    mixed[1] = "I am an int"
    mixed[2.0] = "I am a float"

    assert mixed.get("STRING") == "I am a string"
    assert mixed.get(1) == "I am an int"
    assert mixed.get(2.0) == "I am a float"

    assert mixed.pop("STRING") == "I am a string"
    assert mixed.pop(2.0) == "I am a float"

    assert len(mixed) == 1

    assert not mixed == x
    assert mixed == mixed

    return


"""
Copyright 2022 Veza Technologies Inc.

Use of this source code is governed by the MIT
license that can be found in the LICENSE file or at
https://opensource.org/licenses/MIT.
"""

import pytest
from oaaclient.templates import Tag, OAATemplateException


def test_tag_init():
    new_tag = Tag("test", "value")
    assert new_tag is not None
    assert new_tag.key == "test"
    assert new_tag.value == "value"


def test_tag_characters():

    good_tag = Tag("lettér123_", "letter.123@something.com,characters456@blah.com and space")
    assert good_tag.key == "lettér123_"
    assert good_tag.value == "letter.123@something.com,characters456@blah.com and space"

    with pytest.raises(OAATemplateException) as e:
        Tag("illegal:value!")

    assert e is not None
    assert "Invalid characters in tag key" in e.value.message

    with pytest.raises(OAATemplateException) as e:
        Tag("goodkey", "bad!value*")

    assert e is not None
    assert "Invalid characters in tag value" in e.value.message


def test_tag_equality():

    tag_a = Tag("tagA", "valueA")
    tag_a_same = Tag("tagA", "valueA")
    tag_a_different = Tag("tagA")

    tag_b = Tag("tagB", "valueB")

    assert tag_a == tag_a_same
    assert tag_a != tag_a_different
    assert tag_a != tag_b



"""
Copyright 2022 Veza Technologies Inc.

Use of this source code is governed by the MIT
license that can be found in the LICENSE file or at
https://opensource.org/licenses/MIT.
"""

import base64
import logging
import os
import time
import uuid
import urllib3
import io
from http.client import HTTPMessage, HTTPResponse
from unittest.mock import MagicMock, PropertyMock, patch

import pytest
import requests
from oaaclient.client import OAAClient, OAAClientError, OAAConnectionError, OAAResponseError
from requests.models import Response

from generate_app import generate_app
from generate_idp import generate_idp

# enable debug logging
logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger()
log.setLevel(logging.DEBUG)


@pytest.mark.skipif(not os.getenv("PYTEST_VEZA_HOST"), reason="Test host is not configured")
def test_client_provider(veza_con):
    """ tests for client provider management code using live API

    Does not use the app_provider fixture since this includes all the additional validations
    and tests around provider create/delete behavior
    """
    test_uuid = uuid.uuid4()
    provider_name = f"Pytest-{test_uuid}"

    all_providers = veza_con.get_provider_list()
    assert isinstance(all_providers, list)

    # ensure that randomly generated provider doesn't already exist
    provider_exists = veza_con.get_provider(provider_name)
    assert provider_exists is None

    created_provider = veza_con.create_provider(provider_name, "application")
    assert created_provider is not None
    assert created_provider.get("name") == provider_name
    assert created_provider.get("custom_template") == "application"
    assert created_provider.get("state") == "ENABLED"
    assert created_provider.get("id") is not None

    provider_info = veza_con.get_provider_by_id(created_provider.get("id"))
    assert provider_info is not None
    assert provider_info.get("id") == created_provider.get("id")
    assert provider_info.get("name") == provider_name

    # test getting non-existed
    not_provider = veza_con.get_provider_by_id(uuid.uuid4())
    assert not_provider is None

    # delete the provider
    delete_response = veza_con.delete_provider(created_provider["id"])
    assert isinstance(delete_response, dict)

    deleted = False
    deleted_provider = veza_con.get_provider(provider_name)
    print(deleted_provider)
    if not deleted_provider:
        # delete already succeeded
        deleted = True
    elif deleted_provider["state"] == "DELETING":
        deleted = True
    print(deleted_provider)
    assert deleted

@pytest.mark.skipif(not os.getenv("PYTEST_VEZA_HOST"), reason="Test host is not configured")
def test_client_data_source(veza_con):
    """ tests for client data source management code using live API """

    test_uuid = uuid.uuid4()
    provider_name = f"Pytest-{test_uuid}"
    created_provider = veza_con.create_provider(provider_name, "application")
    assert created_provider is not None
    assert created_provider.get("id") is not None

    provider_id = created_provider["id"]
    existing_data_sources = veza_con.get_data_sources(provider_id)
    # newly created provider should have no data sources
    assert existing_data_sources == []

    not_created = veza_con.get_data_source(name="not created", provider_id=provider_id)
    # expect none for a data source we know doesn't exist yet
    assert not_created is None

    data_source_1 = veza_con.create_data_source(name="data source 1", provider_id=provider_id)
    assert data_source_1 is not None
    assert data_source_1.get("name") == "data source 1"
    assert data_source_1.get("id") is not None

    data_source_2 = veza_con.create_data_source(name="data source 2", provider_id=provider_id)
    assert data_source_2 is not None
    assert data_source_2.get("name") == "data source 2"
    assert data_source_2.get("id") is not None

    assert data_source_1["id"] != data_source_2["id"]

    data_source_1_info = veza_con.get_data_source(name="data source 1", provider_id=provider_id)
    assert data_source_1_info is not None
    assert data_source_1_info.get("name") == "data source 1"
    assert data_source_1_info.get("status") is not None
    assert data_source_1_info.get("id") is not None

    data_source_list = veza_con.get_data_sources(provider_id)
    assert len(data_source_list) == 2

    # test delete
    delete_response = veza_con.delete_data_source(data_source_id=data_source_1["id"], provider_id=provider_id)
    assert isinstance(delete_response, dict)

    deleted = False
    deleted_datasource = veza_con.get_data_source(name="data source 1", provider_id=provider_id)
    if not deleted_datasource:
        # delete already succeeded
        deleted = True
    elif deleted_datasource["status"] == "DELETING":
        deleted = True

    assert deleted

    # delete provider for cleanup
    veza_con.delete_provider(provider_id)


@pytest.mark.parametrize("url",["https://noreply.vezacloud.com", "noreply.vezacloud.com", "noreply.vezacloud.com/", "https://noreply.vezacloud.com/"])
def test_url_formatter(url):
    test_api_key = "1234"
    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

        assert veza_con.url == "https://noreply.vezacloud.com"

@patch.object(requests.Session, "request")
def test_api_get_error(mock_session_get):
    # Test that the correct OAAClient exception is raised on properly populated

    test_api_key = "1234"
    # patch _test_connection to instantiate a connection object
    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url="https://noreply.vezacloud.com", token=test_api_key)


    mock_response = Response()
    mock_response.status_code = 400
    error_message = b"""
                    {
                        "code": "Internal",
                        "message": "Internal Server Error, please retry and if the error persists contact support at support@veza.com",
                        "request_id": "2271c08a9abd3b425c88a397b01bb351",
                        "timestamp": "2022-08-05T21:12:29.405153171Z",
                        "details": [
                            {
                                "@type": "type.googleapis.com/errorstatus.v1.UserFacingErrorInfo",
                                "reason": "INTERNAL",
                                "metadata": {},
                                "message": "Internal server error.",
                                "resolution": "Please retry and if the error persists contact support at support@veza.com"
                            }
                        ]
                    }
                    """
    mock_response._content = error_message
    mock_session_get.return_value = mock_response

    with pytest.raises(OAAClientError) as e:
        response = veza_con.api_get("/api/path")

    # test that the error is populated property
    assert e.value.error == "Internal"
    assert e.value.message == "Internal Server Error, please retry and if the error persists contact support at support@veza.com"
    assert e.value.status_code == 400
    assert len(e.value.details) == 1
    assert "Internal server error." in str(e.value.details)

@patch("urllib3.connectionpool.HTTPConnectionPool._make_request")
@patch("time.sleep", return_value=None)
def test_api_get_retry(mock_sleep, mock_make_request):
    # test that the retry logic behaves correctly, that it retries the right number of times and that total time is within our expected
    test_api_key = "1234"
    # patch _test_connection to instantiate a connection object
    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url="https://noreply.vezacloud.com", token=test_api_key)

    url = "/api/should/fail"

    bad = urllib3.response.HTTPResponse(status=500, reason="Failure", request_url=url)
    mock_make_request.return_value = bad

    with pytest.raises(OAAConnectionError) as e:
        response = veza_con.api_get(url)

    # ten retries (default) would require nine sleep back off calls
    assert mock_sleep.call_count == 9
    sleep_times = [c.args[0] for c in mock_sleep.call_args_list]
    assert sum(sleep_times) == 157.2

    # test that the error is populated property
    assert e.value.error == "ERROR"
    assert "Max retries exceeded" in e.value.message

    # test that retry to good works correct, two bad responses and a good
    url = "/api/should/work"

    mock_make_request.reset_mock()

    response_body = io.BytesIO(b'{"message": "ok"}')
    headers={"Content-Type": "application/json", "Content-Length": "17"}
    good = urllib3.response.HTTPResponse(response_body, status=200, request_url=url, headers=headers, preload_content=False)

    # two bad responses then a good should result in no exceptions raised to caller
    mock_make_request.side_effect = [bad, bad, good]

    # test that the api get will retry to get to the good response
    response = veza_con.api_get(url)
    log.debug(response)
    assert response.get("message") == "ok"

    return

@patch.object(requests.Session, "request")
def test_api_get_nonjson_error(mock_session_get):
    # Test that the OAAClient correctly handles a non-JSON respponse if error isn't coming from Veza stack

    test_api_key = "1234"
    url = "https://noreply.vezacloud.com"
    # patch _test_connection to instantiate a connection object
    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

    # Mock a response with non-JSON data, will force a JSONDecodeError
    mock_response = Response()
    mock_response.status_code = 500
    mock_response._content = b"This is not json"
    mock_response.reason = "Error Reason"
    mock_response.url = url

    mock_session_get.return_value = mock_response

    with pytest.raises(OAAClientError) as e:
        veza_con.api_get("/api/path")

    # should recieve the generic error message
    assert e.value.error == "ERROR"
    assert "Error Reason" in e.value.message
    assert e.value.status_code == 500

@patch.object(requests.Session, "request")
def test_api_get_nonjson_success(mock_session_get):
    # Test that the OAAClient raises the correct error if somehow it gets a successful HTTP response that is not JSON

    test_api_key = "1234"
    url = "https://noreply.vezacloud.com"
    # patch _test_connection to instantiate a connection object
    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

    # Mock a response with non-JSON data, will force a JSONDecodeError
    mock_response = Response()
    mock_response.status_code = 200
    mock_response._content = b"This is not json"
    mock_response.url = url

    mock_session_get.return_value = mock_response

    with pytest.raises(OAAClientError) as e:
        veza_con.api_get("/api/path")

    # should receive the generic error message
    assert e.value.error == "ERROR"
    assert e.value.message == "Response not JSON"
    assert e.value.status_code == 200

@patch.object(requests.Session, "request")
def test_api_post_error(mock_session_post):
    # Test that the correct OAAClient exception is raised on properly populated

    test_api_key = "1234"
    # patch _test_connection to instantiate a connection object
    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url="https://noreply.vezacloud.com", token=test_api_key)


    mock_response = Response()
    mock_response.status_code = 400
    error_message = b"""
            {
                "code": "InvalidArgument",
                "message": "Invalid Arguments",
                "request_id": "1091d23a67ad44a63723fc050280e5ae",
                "timestamp": "2022-08-05T19:59:11.508388808Z",
                "details": [
                    {
                    "@type": "type.googleapis.com/google.rpc.BadRequest",
                    "field_violations": [
                        {
                        "field": "name",
                        "description": "Provider with the same name already exists"
                        }
                    ]
                    },
                    {
                    "@type": "type.googleapis.com/errorstatus.v1.UserFacingErrorInfo",
                    "reason": "INVALID_ARGUMENTS",
                    "metadata": {},
                    "message": "Request includes invalid arguments.",
                    "resolution": "Reference error details for the exact field violations."
                    }
                ]
            }
            """
    mock_response._content = error_message

    mock_session_post.return_value = mock_response

    with pytest.raises(OAAClientError) as e:
        veza_con.api_post("/api/path", data={})

    # test that the error is populated propery
    assert e.value.error == "InvalidArgument"
    assert e.value.message == "Invalid Arguments"
    assert e.value.status_code == 400
    assert e.value.details != []
    assert "Provider with the same name already exists" in str(e.value.details)

@patch.object(requests.Session, "request")
def test_api_post_nonjson_error(mock_session_post):
    # Test that the OAAClient correctly handles a non-JSON respponse if error isn't coming from Veza stack

    test_api_key = "1234"
    url = "https://noreply.vezacloud.com"
    # patch _test_connection to instantiate a connection object
    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

    # Mock a response with non-JSON data, will force a JSONDecodeError
    mock_response = Response()
    mock_response.status_code = 500
    mock_response._content = b"This is not json"
    mock_response.reason = "Error Reason"
    mock_response.url = url

    mock_session_post.return_value = mock_response

    with pytest.raises(OAAClientError) as e:
        veza_con.api_post("/api/path", data={})

    # should recieve the generic error message
    assert e.value.error == "ERROR"
    assert "Error Reason" in e.value.message
    assert e.value.status_code == 500

@patch.object(requests.Session, "request")
def test_api_delete_error(mock_session_delete):
    # Test that the correct OAAClient exception is raised on properly populated

    test_api_key = "1234"
    # patch _test_connection to instantiate a connection object
    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url="https://noreply.vezacloud.com", token=test_api_key)


    mock_response = Response()
    mock_response.status_code = 404
    error_message = b"""
            {
                "code": "NotFound",
                "message": "Not Found",
                "request_id": "1de5e43499c90f2036cdfe92ed76f58e",
                "timestamp": "2022-08-05T21:06:53.046972349Z",
                "details": [
                    {
                    "@type": "type.googleapis.com/errorstatus.v1.ResourceInfo",
                    "resource_type": "datasource",
                    "resource": "b1e654e7-2104-4180-9dee-2f76e2b52463"
                    },
                    {
                    "@type": "type.googleapis.com/errorstatus.v1.UserFacingErrorInfo",
                    "reason": "NOT_FOUND",
                    "metadata": {},
                    "message": "Requested resource was not found.",
                    "resolution": ""
                    }
                ]
            }
            """
    mock_response._content = error_message
    mock_session_delete.return_value = mock_response

    with pytest.raises(OAAClientError) as e:
        veza_con.api_delete("/api/path")

    # test that the error is populated propery
    assert e.value.error == "NotFound"
    assert e.value.message == "Not Found"
    assert e.value.status_code == 404
    assert e.value.details != []
    assert "Requested resource was not found." in str(e.value.details)

@patch.object(requests.Session, "request")
def test_api_post_delete_error(mock_session_delete):
    # Test that the OAAClient correctly handles a non-JSON response if error isn't coming from Veza stack

    test_api_key = "1234"
    url = "https://noreply.vezacloud.com"
    # patch _test_connection to instantiate a connection object
    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

    # Mock a response with non-JSON data, will force a JSONDecodeError
    mock_response = Response()
    mock_response.status_code = 500
    mock_response._content = b"This is not json"
    mock_response.reason = "Error Reason"
    mock_response.url = url

    mock_session_delete.return_value = mock_response

    with pytest.raises(OAAClientError) as e:
        veza_con.api_delete("/api/path")

    # should receive the generic error message
    assert e.value.error == "ERROR"
    assert "Error Reason" in e.value.message
    assert e.value.status_code == 500


@patch('oaaclient.client.requests')
@patch.object(OAAClient, "get_provider", return_value={"id": "123"})
@patch.object(OAAClient, "get_data_source", return_value={"id": "123"})
def test_large_payload(mock_requests, mock_get_provider, mock_get_data_source):
    """Test large payload exception

    Assert that a payload that would be larger than 100MB will throw an exception

    """
    test_api_key = "1234"
    url = "https://noreply.vezacloud.com"

    mock_response = Response()
    mock_response.status_code = 200
    mock_response._content = b"""{"id": "123"}"""
    mock_response.url = url

    mock_requests.get.return_value = mock_response
    mock_requests.post.return_value = mock_response

    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

    # disable compression to make it easier to create a large payload
    veza_con.enable_compression = False

    big = "=" * 100_000_001
    payload = {"data": big}
    with pytest.raises(OAAClientError) as e:
        veza_con.push_metadata("provider_name", "data_source_name", metadata=payload, save_json=False)

    assert e.value.error == "OVERSIZE"
    assert "Payload size exceeds maximum size of 100MB" in e.value.message


@patch('oaaclient.client.requests')
@patch.object(OAAClient, "get_provider", return_value={"id": "123"})
@patch.object(OAAClient, "get_data_source", return_value={"id": "123"})
def test_compression(mock_requests, mock_get_provider, mock_get_data_source):
    """Test large payload exception

    Assert that a payload that would be larger than 100MB will throw an exception

    """
    test_api_key = "1234"
    url = "https://noreply.vezacloud.com"

    mock_response = Response()
    mock_response.status_code = 200
    mock_response._content = b"""{"id": "123"}"""
    mock_response.url = url

    mock_requests.get.return_value = mock_response
    mock_requests.post.return_value = mock_response

    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

    veza_con.enable_compression = True

    app = generate_app()

    with patch.object(veza_con, "api_post") as post_mock:
        veza_con.push_application("provider_name", "data_source_name", application_object=app, save_json=False)

    assert post_mock.called
    call = post_mock.mock_calls[0]

    # get the payload that was posted
    payload = call.args[1]
    # assert compression_type is set in the payload correctly
    assert payload['compression_type'] == "GZIP"
    # assert that the payload is base64 encoded by trying to decode
    assert base64.b64decode(payload['json_data'])


@patch.object(requests.Session, "request")
def test_request_exceptions(mock_session):

    test_api_key = "1234"
    url = "https://noreply.vezacloud.com"
    # patch _test_connection to instantiate a connection object
    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

    # Mock a response with non-JSON data, will force a JSONDecodeError
    mock_response = Response()
    mock_response.status_code = 401
    mock_response._content = b"""
    {
        "code": "InvalidArgument",
        "message": "Invalid Arguments",
        "request_id": "171d043ee1f2bd1ed6f2881f5fc4c505",
        "timestamp": "2022-08-05T19:33:38.838666103Z",
        "details": [
            {
            "@type": "type.googleapis.com/google.rpc.BadRequest",
            "field_violations": [
                {
                "field": "identity_to_permissions.role_assignments.role",
                "description": "Can't connect identity to role as role not found for application (application: SampleApp, role: Administrator, identity: my_user)"
                }
            ]
            },
            {
            "@type": "type.googleapis.com/errorstatus.v1.UserFacingErrorInfo",
            "reason": "INVALID_ARGUMENTS",
            "metadata": {},
            "message": "Request includes invalid arguments.",
            "resolution": "Reference error details for the exact field violations."
            }
        ]
    }
    """
    mock_response.reason = "InvalidArgument"
    mock_response.url = "https://noreply.vezacloud.com/api/v1/call"

    mock_session.return_value = mock_response

    with pytest.raises(OAAResponseError) as e:
        veza_con.api_delete("/api/path")

    # should receive the generic error message
    assert e.value.error == "InvalidArgument"
    assert e.value.message == "Invalid Arguments"
    assert e.value.status_code == 401
    assert e.value.timestamp == "2022-08-05T19:33:38.838666103Z"
    assert e.value.request_id == "171d043ee1f2bd1ed6f2881f5fc4c505"
    assert isinstance(e.value.details, list)
    assert len(e.value.details) == 2

    return

def test_exception_hierarchy():
    """Test exception hierarchy

    Ensure that detailed exceptions are captured by the OAAClientError base exception
    """

    with pytest.raises(OAAClientError) as e:
        raise OAAResponseError("test error", message="test")

    assert e is not None
    assert isinstance(e.value, OAAClientError)
    assert isinstance(e.value, OAAResponseError)

    with pytest.raises(OAAClientError) as e:
        raise OAAConnectionError("connection error", message="test")

    assert e is not None
    assert isinstance(e.value, OAAClientError)
    assert isinstance(e.value, OAAConnectionError)


def test_connection_test():
    """Ensure that connection test fails fast with a bad URL

    Test should successfully fail faster than the timeout
    """

    with pytest.raises(OAAConnectionError) as e:
        con = OAAClient(url="https://host.invalid.com", api_key="test123=")

    assert isinstance(e.value, OAAConnectionError)
    # ensure that the connection test failed because of the DNS error
    assert e.value.error == "Unknown host"

@pytest.mark.skipif(not os.getenv("PYTEST_VEZA_HOST"), reason="Test host is not configured")
@pytest.mark.timeout(10)
def test_bad_api_key():
    """Ensure test fails with invalid API key

    Test should successfully fail faster than timeout
    """

    test_deployment = os.getenv("PYTEST_VEZA_HOST", "")
    test_api_key = os.getenv("VEZA_API_KEY", "")

    decoded = base64.b64decode(test_api_key)
    mangled = str(decoded[:4]) + "XXXX"
    test_api_key = base64.b64encode(decoded)
    # test_api_key = test_api_key.replace(test_api_key[:4], "XXXX")

    log.debug(f"Test bad API key value: {test_api_key=}")
    with pytest.raises(OAAClientError) as e:
        con = OAAClient(url=test_deployment, api_key=test_api_key)

    assert isinstance(e.value, OAAClientError)
    assert e.value.status_code == 401


@patch.object(requests.Session, "request")
def test_api_paging_get(mock_request):

    test_api_key = "1234"
    url = "https://noreply.vezacloud.com"

    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

    # build some fake pages
    page_1 = b"""{"values": ["1", "2", "3"], "has_more": true, "next_page_token": "page2"}"""
    page_2 = b"""{"values": ["4", "5", "6"], "has_more": true, "next_page_token": "page3"}"""
    page_3 = b"""{"values": ["7", "8"], "has_more": false}"""

    responses = []

    for page in [page_1, page_2, page_3]:
        mock_response = Response()
        mock_response.status_code = 200
        mock_response._content = page
        mock_response.url = url

        responses.append(mock_response)

    mock_request.side_effect = responses

    providers = veza_con.get_provider_list()

    assert providers == ["1", "2", "3", "4", "5", "6", "7", "8"]

    # assert that it made three request calls to get through all the pages
    assert mock_request.call_count == 3

    # check that the page_token parame is set correctly on each of the three expected API calls
    call_1 = mock_request.call_args_list[0]
    assert "page_size" in call_1.kwargs.get("params")

    call_2 = mock_request.call_args_list[1]
    assert "page_token=page2" in call_2.kwargs.get("params")
    assert "page_size" in call_2.kwargs.get("params")

    call_3 = mock_request.call_args_list[0]
    assert "page_size" in call_3.kwargs.get("params")

@patch.object(requests.Session, "request")
def test_api_paging_post(mock_request):

    test_api_key = "1234"
    url = "https://noreply.vezacloud.com"

    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

    # build some fake pages
    page_1 = b"""{"values": ["1", "2", "3"], "has_more": true, "next_page_token": "page2"}"""
    page_2 = b"""{"values": ["4", "5", "6"], "has_more": true, "next_page_token": "page3"}"""
    page_3 = b"""{"values": ["7", "8"], "has_more": false}"""

    responses = []

    for page in [page_1, page_2, page_3]:
        mock_response = Response()
        mock_response.status_code = 200
        mock_response._content = page
        mock_response.url = url

        responses.append(mock_response)

    mock_request.side_effect = responses

    result = veza_con.api_post("/fake/url", data={})

    assert result == ["1", "2", "3", "4", "5", "6", "7", "8"]

    # assert that it made three request calls to get through all the pages
    assert mock_request.call_count == 3

    # check that the page_token params is set correctly on each of the three expected API calls
    # first call should have no page size
    call_1 = mock_request.call_args_list[0]
    assert call_1.kwargs.get("params") is None

    call_2 = mock_request.call_args_list[1]
    assert "page_token=page2" in call_2.kwargs.get("params")

    call_3 = mock_request.call_args_list[2]
    assert "page_token=page3" in call_3.kwargs.get("params")

@patch.object(requests.Session, "request")
def test_api_paging_put(mock_request):

    test_api_key = "1234"
    url = "https://noreply.vezacloud.com"

    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

    # build some fake pages
    page_1 = b"""{"values": ["1", "2", "3"], "has_more": true, "next_page_token": "page2"}"""
    page_2 = b"""{"values": ["4", "5", "6"], "has_more": true, "next_page_token": "page3"}"""
    page_3 = b"""{"values": ["7", "8"], "has_more": false}"""

    responses = []

    for page in [page_1, page_2, page_3]:
        mock_response = Response()
        mock_response.status_code = 200
        mock_response._content = page
        mock_response.url = url

        responses.append(mock_response)

    mock_request.side_effect = responses

    result = veza_con.api_put("/fake/url", data={})

    assert result == ["1", "2", "3", "4", "5", "6", "7", "8"]

    # assert that it made three request calls to get through all the pages
    assert mock_request.call_count == 3

    # check that the page_token params is set correctly on each of the three expected API calls
    # first call should have no page size
    call_1 = mock_request.call_args_list[0]
    assert call_1.kwargs.get("params") is None

    call_2 = mock_request.call_args_list[1]
    assert "page_token=page2" in call_2.kwargs.get("params")

    call_3 = mock_request.call_args_list[2]
    assert "page_token=page3" in call_3.kwargs.get("params")


def test_allowed_characters():

    test_api_key = "1234"
    url = "https://noreply.vezacloud.com"

    with patch.object(OAAClient, "_test_connection", return_value=None):
        veza_con = OAAClient(url=url, token=test_api_key)

    with pytest.raises(ValueError) as e:
        veza_con.create_provider("invalid/characters", "application")

    assert e.value is not None
    assert "Provider name contains invalid characters" in str(e.value)

    with pytest.raises(ValueError) as e:
        veza_con.create_data_source("invalid/characters", provider_id="1234")

    assert e.value is not None
    assert "Data source name contains invalid characters" in str(e.value)

    with patch.object(OAAClient, "api_post", return_value={}):
        provider = veza_con.create_provider("allowed 1234 @#$%&*:()!,_'\" =.-", "application")

    assert provider == {}


@pytest.mark.skipif(not os.getenv("PYTEST_VEZA_HOST"), reason="Test host is not configured")
def test_create_query(veza_con):
    """Test the methods for managing queries """

    existing_queries = veza_con.get_queries()

    assert isinstance(existing_queries, list)
    starting_count = len(existing_queries)

    query_name_uuid = uuid.uuid4()
    test_query = {
                    "name": f"Pytest test query {query_name_uuid}",
                    "description": "Pytest Generated",
                    "category": "IDP_ANALYSIS",
                    "level": "BASIC",
                    "result_type": "NUMBER",
                    "query_type": "SOURCE_TO_DESTINATION",
                    "source_node_types": {
                        "nodes": [
                            {
                                "node_type": "OktaUser",
                                "tags": [],
                                "conditions": [],
                                "node_id": "",
                                "excluded_tags": [],
                                "count_conditions": []
                            }
                        ],
                        "nodes_operator": "AND"
                    }
                }

    create_response = veza_con.create_query(test_query)
    assert isinstance(create_response, dict)
    created_id = create_response["id"]
    assert created_id is not None

    get_response = veza_con.get_query_by_id(created_id)
    assert isinstance(get_response, dict)
    assert get_response['id'] == created_id

    delete_response = veza_con.delete_query(created_id)
    assert isinstance(delete_response, dict)

    with pytest.raises(OAAResponseError) as e:
        veza_con.get_query_by_id(created_id)

    assert e.value is not None
    assert e.value.status_code == 404

@pytest.mark.skipif(not os.getenv("PYTEST_VEZA_HOST"), reason="Test host is not configured")
def test_create_report(veza_con):
    """Test the methods for managing reports """

    existing_reports = veza_con.get_reports()
    assert isinstance(existing_reports, list)
    starting_count = len(existing_reports)

    report_name_uuid = uuid.uuid4()

    report_definition = { "name": f"Pytest Report {report_name_uuid}", "description": "Created for Pytest", "category": "OAA", "queries": []}

    create_response = veza_con.create_report(report=report_definition)
    assert isinstance(create_response, dict)
    created_id = create_response.get("id")
    assert created_id is not None

    get_response = veza_con.get_report_by_id(id=created_id)
    assert isinstance(get_response, dict)
    assert get_response["id"] == created_id
    assert len(get_response["queries"]) == 0

    query_name_uuid = uuid.uuid4()
    test_query = {
                    "name": f"Pytest test query {query_name_uuid}",
                    "description": "Pytest Generated",
                    "category": "IDP_ANALYSIS",
                    "level": "BASIC",
                    "result_type": "NUMBER",
                    "query_type": "SOURCE_TO_DESTINATION",
                    "source_node_types": {
                        "nodes": [
                            {
                                "node_type": "OktaUser",
                                "tags": [],
                                "conditions": [],
                                "node_id": "",
                                "excluded_tags": [],
                                "count_conditions": []
                            }
                        ],
                        "nodes_operator": "AND"
                    }
                }

    query_create_response = veza_con.create_query(test_query)
    query_id = query_create_response["id"]

    add_response = veza_con.add_query_report(report_id=created_id, query_id=query_id)
    log.debug(add_response)
    assert len(add_response["queries"]) == 1

    add_response = veza_con.add_query_report(report_id=created_id, query_id=query_id)
    log.debug(add_response)

    assert len(add_response["queries"]) == 1

    get_response = veza_con.get_report_by_id(id=created_id, include_inactive_queries=True)
    assert isinstance(get_response, dict)
    assert get_response["id"] == created_id
    log.debug(get_response)
    assert len(get_response["queries"]) == 1

    get_response["name"] = f"Updated Pytest {query_name_uuid}"
    update_response = veza_con.update_report(report_id=created_id, report=get_response)
    log.debug(update_response)
    assert update_response["name"] == f"Updated Pytest {query_name_uuid}"

    delete_response = veza_con.delete_report(created_id)
    assert isinstance(delete_response, dict)

    with pytest.raises(OAAResponseError) as e:
        veza_con.get_report_by_id(created_id)

    veza_con.delete_query(query_id)